/*!
\file skymap.cpp
\brief Class implementation for Skymap
\author Nicholas Phillips, phillips@friday.gsfc.nasa.gov
\date July 21, 2003
*/
#include <fstream>
#include <iostream>
#include <iomanip>
#include <math.h>
#include <fitsio.h>
#include "skymap.h"

using namespace std;

const double r2d = 180/M_PI; // radians to degrees

//================================================
unsigned int nrings(unsigned int nside_) { return 4*(nside_)-1; }
/*
  \c region = 0 if North cap, 1 if Equalatorial, 2 if South cap
*/
unsigned int num_per_ring(unsigned int nside_, unsigned int iring, int & region) {
	if( iring < nside_ ) {
		region = 0;
		return 4*(iring+1);
	}
	else if( iring < 3*nside_ ) {
		region = 1;
		return 4*nside_;
	}
	region = 2;
	return 4*nside_ - 4*(iring - 3*nside_ + 1);
}

//==============================================================
Skymap::Skymap() {
	nside_ = 0;
	npixels_ = 0;
	ring_order = true;
	has_polar = false;
} 


Skymap::~Skymap() {
}

//==============================================================================================
int Skymap::nside(const int nside__) {
	const int valid_nside[] = {1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,-1};
	unsigned int i;
	for(i =0; valid_nside[i] != -1; i++)
		if( nside__ == valid_nside[i] ) break;
	if( valid_nside[i] == -1 ) {
		cerr << "Skymap::nside:Invalid value, " << nside__ << ". Nothing done" << endl;
		return nside_;
	}
	if( nside__ == nside_ ) {
		cerr << "Skymap::nside:Warning: same value, nothing done." << endl;
		return nside_;
	}
	nside_ = nside__;
	npixels_ = nside2npix(nside_);
	resize(npixels_);
	for(i = 0; i < size(); i++ ) (*this)[i].I = (*this)[i].Q = (*this)[i].U = (*this)[i].Nobs = 0;
	//has_polar = has_Nobs_ = false;
	cout << "Skymap::nside:Resized to " << nside() << ",  " << npixels_ << " pixels." << endl;
	return nside_;
}
//----------------------------------------------------------------------------------------------
int Skymap::nside(const int nside__, bool ring_order_) {
	ring_order = ring_order_;
	return nside(nside__);
}

//====================================
bool Skymap::loadfitsfile(string filename) {
	fitsfile *fptr;                       // FITS file pointer, defined in fitsio.h 
	int status = 0;                   // CFITSIO status value MUST be initialized to zero! 

	// Try to open file and prepare to read tables
	fits_open_file(&fptr, filename.c_str(), READONLY, &status);
	int hdutype;
	fits_movabs_hdu(fptr, 2, &hdutype, &status);
	if ( status ) {
		fits_report_error(stderr, status);
		return false;
	}
	cout << "Skymap::loadfitsfile:Using file '" << filename << "'" << endl;
	if (hdutype != BINARY_TBL) {
		cerr << "Skymap::loadfitsfile: Extension not binary table" << endl;
		return false;
	}

	// Get number of rows and columns, check ncols == 3
	long nrows;
	int ncols;
	fits_get_num_rows(fptr,&nrows,&status);
	fits_get_num_cols(fptr,&ncols,&status);

	// Look for needed KEYWORDS and column numbers
	char comment[FLEN_COMMENT];
	char pixtype[FLEN_VALUE];
	char ordering[FLEN_VALUE];
	fits_read_key(fptr, TSTRING, "PIXTYPE", pixtype, comment, &status);
	fits_read_key(fptr, TSTRING, "ORDERING", ordering, comment, &status);
	if( status ) {
		fits_report_error(stderr, status);
		fits_close_file(fptr, &status);
		return false;
	}
	if( string(pixtype) != "HEALPIX" ) {
		cerr << "Skymap::loadfitsfile:Wrong pixtype" << endl;
		fits_close_file(fptr, &status);
		return false;
	}
	if( string(ordering) == "RING" ) {
		ring_order = true;
	}
	else if( string(ordering) == "NESTED" ) {
		ring_order = false;
	}
	else {
		cerr << "Skymap::loadfitsfile:Wrong Ordering: '" << ordering << "'" << endl;
		fits_close_file(fptr, &status);
		return false;
	}
	cout << "Skymap::loadfitsfile:Map is in " << (ring_order ? "ring" : "nested" ) << " order." << endl;

	// column numbers....
	int I_colnum = 0;
	int Q_colnum = 0;
	int U_colnum = 0;
	int Nobs_colnum = 0;
	has_polar = false;
	has_Nobs_ = false;
	fits_get_colnum(fptr, CASEINSEN, "TEMPERATURE", &I_colnum, &status);
	status = 0;
	fits_get_colnum(fptr, CASEINSEN, "Q-POLARISATION", &Q_colnum, &status);
	if( Q_colnum == 0 ) {
		status = 0;
		fits_get_colnum(fptr, CASEINSEN, "Q_POLARISATION", &Q_colnum, &status);
	}
	status = 0;
	fits_get_colnum(fptr, CASEINSEN, "U-POLARISATION", &U_colnum, &status);
	if( U_colnum == 0 ) {
		status = 0;
		fits_get_colnum(fptr, CASEINSEN, "U_POLARISATION", &U_colnum, &status);
	}
	status = 0;
	fits_get_colnum(fptr, CASEINSEN, "N_OBS   ", &Nobs_colnum, &status);

	if( I_colnum == 0 ) {
		cout << "Skymap::loadfitsfile:No 'TEMPERATURE' column in file.";
		cout << " Assuming first column" << endl;
		I_colnum = 1;
		//fits_close_file(fptr, &status);
		//return false;
	}
	if( (Q_colnum != 0) && (I_colnum != 0) ) {
		cout << "Skymap::loadfitsfile:Found polarization data." << endl;
		has_polar = true;
	}
	if( Nobs_colnum != 0 ) {
		cout << "Skymap::loadfitsfile:Found N_obs data." << endl;
		has_Nobs_ = true;
	}

	status = 0;

	// Get nside parameter and compare with configured value
	long nside_in;
	if ( fits_read_key_lng(fptr, "NSIDE", &nside_in, comment, &status) ) {
		cerr << "Skymap::loadfitsfile: Missing keyword 'NSIDE'" << endl;
		fits_close_file(fptr, &status);
		return false;
	}
	nside(nside_in);
	// figure out the number of rows per column
	if( (npixels_ % nrows) != 0 ) {
		cerr << "Skymap::loadfitsfile:Problem with 'nside' and 'nrows'" << endl;
		fits_close_file(fptr, &status);
		return false;
	}
	long npercol = npixels_/nrows;

	// Get some memory and start reading in  the data
	float *map_in = new float[npixels_];
	float nulval = 0.0;

	//------------ Read the Temperature data ----------------------------
	int anynul = 0;
	for (int irow = 0; irow < nrows; irow++)
		fits_read_col(fptr, TFLOAT, I_colnum, irow+1, 1, npercol, &nulval, &(map_in[irow*npercol]), &anynul, &status);
	if( anynul != 0 ) cerr << "Skymap::Warning: " << anynul << " null values found reading column " << I_colnum << endl;
	float *mptr = map_in;
	for(iterator si = begin(); si != end(); ++si)  si->I = *mptr++;

	if( has_polar ) {
		//------------ Read the Q-polarization data ----------------------------
		anynul = 0;
		for (int irow = 0; irow < nrows; irow++)
			fits_read_col(fptr, TFLOAT, Q_colnum, irow+1, 1, npercol, &nulval, &(map_in[irow*npercol]), &anynul, &status);
		if( anynul != 0 ) cerr << "Skymap::Warning: " << anynul << " null values found reading column " << Q_colnum << endl;
		mptr = map_in;
		for(iterator si = begin(); si != end(); ++si) si->Q = *mptr++;

		//------------ Read the U-polarization data ----------------------------
		anynul = 0;
		for (int irow = 0; irow < nrows; irow++)
			fits_read_col(fptr, TFLOAT, U_colnum, irow+1, 1, npercol, &nulval, &(map_in[irow*npercol]), &anynul, &status);
		if( anynul != 0 ) cerr << "Skymap::Warning: " << anynul << " null values found reading column " << U_colnum << endl;
		mptr = map_in;
		for(iterator si = begin(); si != end(); ++si)  si->U = *mptr++;
	} // if has_polarization
	if( Nobs_colnum ) {
		//------------ Nobs data ----------------------------
		anynul = 0;
		for (int irow = 0; irow < nrows; irow++)
			fits_read_col(fptr, TFLOAT, Nobs_colnum, irow+1, 1, npercol, &nulval, &(map_in[irow*npercol]), &anynul, &status);
		if( anynul != 0 ) cerr << "Skymap::Warning: " << anynul << " null values found reading column " << Nobs_colnum << endl;
		mptr = map_in;
		for(iterator si = begin(); si != end(); ++si) si->Nobs = *mptr++;
	}

	//----------- Close and get ready to go home -----------------------------
	if( status ) {
		cerr << "Skymap::loadfitsfile:While reading in data:" << endl;
		fits_report_error(stderr, status);
		fits_close_file(fptr, &status);
		return false;
	}

	// Done accessing the file, close and free memory
	if ( fits_close_file(fptr, &status) ) {
		fits_report_error(stderr, status);
		return false;
	}
	delete[] map_in;
	// Done, report all ok and go home
	cout << "Skymap::loadfitsfile:Successful load." << endl;
	return true;
}

//============================================
//============================================
/*! Save the skymap as a FITS file. All data is in the second HDU, stored as a
      Binary Table. The table has six columns:
\code
PIXEL  TEMPERATURE  Q-POLARISATION  U-POLARISATION
long      float          float          float
Index    Kelvin         Kelvin         Kelvin
 \endcode
*/
bool Skymap::savefitsfile(string filename) {
	char *extname = { "Skymap" };

	fitsfile *fptr;                       // FITS file pointer, defined in fitsio.h
	int status = 0;                   // CFITSIO status value MUST be initialized to zero!

	// Try to open file and start binary table
	fits_create_file(&fptr, filename.c_str(), &status);
	if( has_polar ) {
		char *ttype[] = { "PIXEL", "TEMPERATURE", "Q-POLARISATION", "U-POLARISATION" };
		char *tform[] = { "1J", "1E", "1E", "1E" };
		char *tunit[] = { "Index", "Kelvin", "Kelvin", "Kelvin" };
		fits_create_tbl(fptr, BINARY_TBL, 0, 4, ttype, tform, tunit, extname, &status);
	}
	else {
		/*
		char *ttype[] = { "PIXEL", "TEMPERATURE" };
		char *tform[] = { "1J", "1E" };
		char *tunit[] = { "Index", "Kelvin" };
		*/
		char *ttype[] = { "TEMPERATURE" };
		char *tform[] = { "1E" };
		char *tunit[] = { "Kelvin" };
		fits_create_tbl(fptr, BINARY_TBL, 0, 1, ttype, tform, tunit, extname, &status);
	}
	fits_write_date(fptr, &status);

	// Add some useful parameters to the header
	char *svalue = "HEALPIX";
	fits_write_key(fptr, TSTRING, "PIXTYPE", svalue, "Healpix pixel scheme", &status);
	long lvalue = nside_;
	fits_write_key(fptr, TLONG, "NSIDE", &lvalue, "Healpix nside parameter", &status);
	svalue = "RING";
	fits_write_key(fptr, TSTRING, "ORDERING", svalue, "Healpix pixel order", &status);
	lvalue = npixels_;
	fits_write_key(fptr, TLONG, "NPIXELS", &lvalue, "Number of observed pixels", &status);
	// Check no errors so far
	if ( status ) {
		cout << "Skymap::savefitsfile:CFITSIO error:" << endl;
		fits_report_error(stderr, status);
		return false;
	}
	cout << "Skymap::savefitsfile:Using file '" << filename << "'" << endl;

	// some memory and iterators for saving
	long *larray = new long[size()];
	float *farray = new float[size()];
	unsigned int i;
	//===================================================
	// Let's do some saving
	/*
	// Pixel Number
	for(i = 0; i < size(); ++i)  larray[i] = i;
	fits_write_colnull(fptr, TLONG, 1, 1, 1, size(), larray, NULL, &status);
	*/
	// I stokes parameter
	for(i = 0; i < size(); ++i)  farray[i] = (*this)[i].I;
	fits_write_colnull(fptr, TFLOAT, 1, 1, 1, size(), farray, NULL, &status);
	if( has_polar ) {
		// Q stokes parameter
		for(i = 0; i < size(); ++i)  farray[i] = (*this)[i].Q;
		fits_write_colnull(fptr, TFLOAT, 3, 1, 1, size(), farray, NULL, &status);
		// U stokes parameter
		for(i = 0; i < size(); ++i)  farray[i] = (*this)[i].U;
		fits_write_colnull(fptr, TFLOAT, 4, 1, 1, size(), farray, NULL, &status);
	}
	// check no problems saving the data
	if ( status ) {
		cout << ", CFITSIO data writing error:" << endl;
		fits_report_error(stderr, status);
		return false;
	}
	//===================================================
	// Done accessing the file, close and free memory
	if ( fits_close_file(fptr, &status) ) {
	cout << ",  CFITSIO close error:" << endl;
	fits_report_error(stderr, status);
	return false;
	}
	delete[] larray;
	delete[] farray;
	cout << "Skymap::savefitsfile:Successful save." << endl;
	return true;
}



//=========================================
void Skymap::zero_mean() {
	cout << "Skymap::zero_mean:Only I adjusted." << endl;
	double mean = 0;
	for(unsigned int i = 0; i < size(); i++)
		mean += (*this)[i].I;
	mean /= size();
	for(unsigned int i = 0; i < size(); i++)
		(*this)[i].I -= mean;
	return;
}



